import ipywidgets as widgets  # type: ignore
from IPython.display import display  # type: ignore
from dataclasses import dataclass, field

OPTION_LIST = {
    "IMAGING TELESCOPE": [
        "Resolution (panchromatic)",
        "Ground sampling distance (panchromatic)",
        "Resolution (multispectral)",
        "Ground sampling distance (multispectral)",
        "Altitude",
        "Half field of view",
        "Mirror aperture",
        "F-number",
        "Focal length",
        "Pixel size (panchromatic)",
        "Pixel size (multispectral)",
        "Swath width",
    ],
    "MITRE_Template": [
        "Object height",
        "Image height",
        "Wavelength",
    ],
}

IMAGING_TELESCOPE = {
    "Resolution (panchromatic)": 1.8,
    "Ground sampling distance (panchromatic)": 0.6,
    "Resolution (multispectral)": 1.8,
    "Ground sampling distance (multispectral)": 0.9,
    "Altitude": 460000,
    "Half field of view": 0.02003638,
    "Mirror aperture": 0.60,
    "F-number": 5.53,
    "Focal length": 3.32,
    "Pixel size (panchromatic)": 8.75e-6,
    "Pixel size (multispectral)": 13e-6,
    "Swath width": 18400,
}

MITRE_Template = {
    "Object height": 1.8,
    "Image height": 1.8,
    "Wavelength": 1.8,
}


@dataclass
class Requirement:
    requirement_name: str
    latex_symbol: str
    value: int
    units: str
    tolerance: float
    sympy_symbol: str = field(init=False)

    def __post_init__(self):
        self.sympy_symbol = self.latex_symbol.replace("{", "").replace("}", "")

    # def getSympySymbol(self):
    #     return latex2sympy(self.latex_symbol)

    @property
    def is_fixed(self):
        return self.tolerance == 0.0

    # @property
    # def equations(self, strict=False):
    #     if self.is_fixed:
    #         return [sympy.Eq(self.getSympySymbol(), self.value)]
    #     else:
    #       signs = [">=", "<="] if not strict else [">", "<"]
    #       bounds = [self.value - self.tolerance, self.value + self.tolerance]
    #       return [
    #             sympy.Rel(self.getSympySymbol(), bound, sign)
    #             for bound, sign in zip(bounds, signs)
    #         ]


def _find_symbol(name, variable_dict):

    matching_keys = [
        key for key, value in variable_dict.items() if name in value["name"]
    ]

    if not matching_keys:
        matching_keys.append("unknown")

    return matching_keys[0]


def requirements_from_table(results, variable_dict):
    requirements = []

    for key, value in results["values"].items():

        latex_symbol = _find_symbol(key, variable_dict)

        name = key
        numerical_value = value["Value"]
        unit = value["Units"]
        tolerance = value["Tolerance"]

        requirements.append(
            Requirement(
                requirement_name=name,
                latex_symbol=latex_symbol,
                value=numerical_value,
                units=unit,
                tolerance=tolerance,
            )
        )

    return requirements


def interactive_table(preset_options_dict, variable_dict):
    """
    Creates an interactive table with a dropdown for selecting options.

    Parameters:
    options_dict (dict): A dictionary where keys are dropdown options and
      values are lists of row names.

    Returns:
    dict: A dictionary containing user inputs for the selected rows.
    """

    variable_names = [details["name"] for details in variable_dict.values()]

    # Placeholder for result dictionary
    result = {}

    # Create dropdown for options
    dropdown = widgets.Dropdown(
        options=list(preset_options_dict.keys()),
        description="Select Option:",
        style={"description_width": "initial"},
    )

    # Dictionary to hold widgets for user input
    value_widgets = {}

    # VBox to stack rows vertically
    rows_output = widgets.VBox()

    # Output widget for confirmation messages
    message_output = widgets.Output()

    # Mutable container to store the current name label width
    name_label_width = ["150px"]  # Default width

    # Function to display the table based on the dropdown selection
    def display_table(change):
        selected_option = change["new"]

        # Clear existing rows
        rows_output.children = []
        value_widgets.clear()

        if selected_option in preset_options_dict:
            rows = preset_options_dict[selected_option]
            max_name_length = max(len(name) for name in rows)
            # Update the name_label_width based on the longest row name
            name_label_width[0] = f"{max_name_length + 2}ch"

            for row_name in rows:
                # Create name label with dynamic width
                name_label = widgets.Label(
                    value=row_name,
                    layout=widgets.Layout(width=name_label_width[0]),
                )

                # Depending on the selected option, set default values
                if selected_option == "IMAGING TELESCOPE":
                    default_value = IMAGING_TELESCOPE.get(row_name, 0.0)
                # elif selected_option == "LIDAR":
                #     default_value = LIDAR.get(row_name, 0.0)
                elif selected_option == "MITRE_Template":
                    default_value = MITRE_Template.get(row_name, 0.0)

                # Create input widgets
                value_text = widgets.FloatText(
                    placeholder="Value",
                    value=default_value,
                    layout=widgets.Layout(width="150px"),
                )
                tolerance_text = widgets.FloatText(
                    placeholder="Tolerance", layout=widgets.Layout(width="150px")
                )
                accuracy_text = widgets.FloatText(
                    placeholder="Accuracy", layout=widgets.Layout(width="150px")
                )
                units_text = widgets.Text(
                    placeholder="Units", layout=widgets.Layout(width="150px")
                )

                # Combine widgets into a horizontal box
                row = widgets.HBox(
                    [
                        name_label,
                        value_text,
                        tolerance_text,
                        accuracy_text,
                        units_text,
                    ]
                )

                # Store the row widgets
                value_widgets[row_name] = row

                # Add the row to the rows_output VBox
                rows_output.children += (row,)

    # Attach handler to dropdown
    dropdown.observe(display_table, names="value")
    display(dropdown)
    display(rows_output)
    display(message_output)

    # Function to collect and store user inputs
    def submit_values(_):
        updated_values = {}

        for key, widget in value_widgets.items():
            variable = widget.children[0].value
            if key.startswith("req_"):
                updated_values[variable] = {
                    "Value": widget.children[1].value,
                    "Tolerance": widget.children[2].value,
                    "Accuracy": widget.children[3].value,
                    "Units": widget.children[4].value,
                }
            else:
                updated_values[key] = {
                    "Value": widget.children[1].value,
                    "Tolerance": widget.children[2].value,
                    "Accuracy": widget.children[3].value,
                    "Units": widget.children[4].value,
                }

        result["values"] = updated_values

        # Display confirmation message
        with message_output:
            message_output.clear_output()

    # Function to add a new requirement row
    def add_req(_):

        unique_key = (
            f"req_{len([k for k in value_widgets if k.startswith('req_')]) + 1}"
        )

        # Create a dropdown for variable selection with dynamic width
        variable_dropdown = widgets.Dropdown(
            options=variable_names,
            description="Variable:",
            style={"description_width": "initial"},
            layout=widgets.Layout(width=2 * name_label_width[0]),
        )
        value_text = widgets.FloatText(
            placeholder="Value",
            layout=widgets.Layout(width="150px"),
        )
        tolerance_text = widgets.FloatText(
            placeholder="Tolerance", layout=widgets.Layout(width="150px")
        )
        accuracy_text = widgets.FloatText(
            placeholder="Accuracy", layout=widgets.Layout(width="150px")
        )
        units_text = widgets.Text(
            placeholder="Units", layout=widgets.Layout(width="150px")
        )

        new_row = widgets.HBox(
            [variable_dropdown, value_text, tolerance_text, accuracy_text, units_text]
        )

        rows_output.children += (new_row,)
        value_widgets[unique_key] = new_row

    submit_button = widgets.Button(description="Submit")
    submit_button.on_click(submit_values)

    add_req_button = widgets.Button(description="Add Requirement")
    add_req_button.on_click(add_req)

    buttons_box = widgets.HBox([submit_button, add_req_button])
    display(buttons_box)

    return result



# This file was auto-generated by Fern from our API Definition.

from .circuit_image import CircuitImage
from .code_synthesis_response import CodeSynthesisResponse
from .edge import Edge
from .extract_links_response import ExtractLinksResponse
from .extract_response import ExtractResponse
from .extract_statements_response import ExtractStatementsResponse
from .extract_title_response import ExtractTitleResponse
from .formalize_response import FormalizeResponse
from .generate_code_response import GenerateCodeResponse
from .http_validation_error import HttpValidationError
from .interactive_response import InteractiveResponse
from .measurement import Measurement
from .netlist import Netlist
from .pic_component import PicComponent
from .pic_component_info_value import PicComponentInfoValue
from .pic_component_settings_value import PicComponentSettingsValue
from .refine_code_response import RefineCodeResponse
from .requirement_body import RequirementBody
from .solution_response import SolutionResponse
from .solution_response_solution_value import SolutionResponseSolutionValue
from .statement import Statement
from .statement_input import StatementInput
from .synthesis_response import SynthesisResponse
from .synthesize_circuit_response import SynthesizeCircuitResponse
from .type import Type
from .validate_response import ValidateResponse
from .validation_error import ValidationError
from .validation_error_loc_item import ValidationErrorLocItem
from .verify_netlist_response import VerifyNetlistResponse
from .verify_response import VerifyResponse

__all__ = [
    "CircuitImage",
    "CodeSynthesisResponse",
    "Edge",
    "ExtractLinksResponse",
    "ExtractResponse",
    "ExtractStatementsResponse",
    "ExtractTitleResponse",
    "FormalizeResponse",
    "GenerateCodeResponse",
    "HttpValidationError",
    "InteractiveResponse",
    "Measurement",
    "Netlist",
    "PicComponent",
    "PicComponentInfoValue",
    "PicComponentSettingsValue",
    "RefineCodeResponse",
    "RequirementBody",
    "SolutionResponse",
    "SolutionResponseSolutionValue",
    "Statement",
    "StatementInput",
    "SynthesisResponse",
    "SynthesizeCircuitResponse",
    "Type",
    "ValidateResponse",
    "ValidationError",
    "ValidationErrorLocItem",
    "VerifyNetlistResponse",
    "VerifyResponse",
]



from IPython import get_ipython  # type: ignore
from IPython.core.magic import register_line_cell_magic, register_line_magic  # type: ignore
from IPython.display import HTML, display  # type: ignore

import platformdirs  # type: ignore
import os
import sys
from . import Axiomatic


class AXMagic:
    """Class implementing magic functions for IPython.
    Import with `%load_ext axiomatic.magic`."""

    def __init__(self):
        self.folder = platformdirs.user_config_dir("axiomatic")
        if not os.path.exists(f"{self.folder}/api_key"):
            os.makedirs(self.folder, exist_ok=True)
            self.api_key = None
        else:
            with open(f"{self.folder}/api_key", "r") as f:
                self.api_key = f.read()
            self.ax = Axiomatic(api_key=self.api_key)
        self.query = ""

    def ax_api(self, query):
        folder = platformdirs.user_config_dir("axiomatic")
        with open(f"{folder}/api_key", "w") as f:
            f.write(query)
            self.api = query
            self.ax = Axiomatic(api_key=self.api)
            print("API key set.")

    def axquery(self, query, cell=None):
        if self.api_key:
            if cell:
                # REFINE
                try:
                    exec(cell)
                    feedback = "Code executed successfully."
                except Exception as e:
                    feedback = f"Errors:\n{e}"
                print(feedback)
                current_query = f"{self.query}\n{query}"
                result = self.ax.pic.refine(
                    query=current_query, code=cell, feedback=feedback
                )
            else:
                # GENERATE FROM SCRATCH
                self.query = query
                result = self.ax.pic.generate(query=query)

            # Process output
            pre_thought = result.raw_content.split("<thought>")[0]
            thought = result.thought_text.replace("\n", "<br>")
            if not thought:
                output = result.raw_content.replace("\n", "<br>")
            else:
                output = pre_thought + "<br>" + thought
            html_content = f"""<div style='font-family: Arial, sans-serif; line-height: 1.5;'>"
<div style='color: #6EB700;'><strong>AX:</strong> {output}</div>"""
            display(HTML(html_content))

            # Process code
            # remove last three lines (saving file)
            if result.code:
                code = "\n".join(result.code.split("\n")[:-3] + ["c"])
                if "google.colab" in sys.modules:
                    # When running in colab
                    from google.colab import _frontend  # type: ignore

                    _frontend.create_scratch_cell(
                        f"""# {query}\n# %%ax_fix\n{code}""", bottom_pane=True
                    )
                else:
                    # When running in jupyter
                    get_ipython().set_next_input(f"# %%ax_fix\n{code}", replace=False)

        else:
            print(
                "Please set your Axiomatic API key first with the command %ax_api API_KEY and restart the kernel. Request the api key at our Customer Service."
            )

    def ax_query(self, query, cell=None):
        # Updates the target circuit query
        self.query = query
        return self.axquery(query, cell)

    def ax_fix(self, query, cell=None):
        # Runs again without updating the query
        return self.axquery(query, cell)


def ax_help(value: str):
    print(
        """
Available commands:

- `%load_ext axiomatic_pic` loads the ipython extension.
- `%ax_api` sets up the API key
- `%ax_query` returns the requested circuit using our experimental API
- `%%ax_fix` edit the given code
"""
    )


def load_ipython_extension(ipython):
    ax_magic = AXMagic()
    ipython.register_magic_function(ax_magic.ax_query, "line_cell")
    ipython.register_magic_function(ax_magic.ax_fix, "line_cell")
    ipython.register_magic_function(ax_magic.ax_api, "line")
    ipython.register_magic_function(ax_help, "line")
